from loguru import logger
from tonclient.types import ParamsOfQueryOperation, ParamsOfBatchQuery, ResultOfQueryCollection

from tvmbase.client import Client
from tvmbase.constants import GRAPHQL_LIMIT
from tvmbase.models.boc_type import BocType
from tvmbase.models.tvm.base import BaseTvm
from tvmbase.types import TvmType


class BatchLoader:

    def __init__(self, client: Client):
        self.client = client

    async def load(self, data: list[tuple[str, BocType]]) -> list[TvmType]:
        assert len(data) <= GRAPHQL_LIMIT, f'Too big batch, size is {len(data)} while max size is {GRAPHQL_LIMIT}'
        logger.debug(f'Loading batch with size {len(data)}...')
        operations = list()
        for idx, boc_type in data:
            tvm_class: BaseTvm = boc_type.value
            query = tvm_class.gql_query(idx)
            operations.append(
                ParamsOfQueryOperation.QueryCollection(params=query)
            )
        params = ParamsOfBatchQuery(operations=operations)
        results = await self.client.net.batch_query(params=params)
        instances = list()
        for (idx, boc_type), result in zip(data, results.results):
            tvm_class: BaseTvm = boc_type.value
            query_result = ResultOfQueryCollection(result)
            instance = await tvm_class.from_query_result(self.client, idx, query_result)
            instances.append(instance)
        return instances



from tonclient.types import ParamsOfParse, ParamsOfQueryCollection, ResultOfQueryCollection

from tvmbase.client import Client
from tvmbase.models.data import AccountData
from tvmbase.models.tvm.base import BaseTvm


class Account(BaseTvm):
    DATA_TYPE: type = AccountData

    def __init__(self, address: str, data: AccountData | None, exists: bool = True):
        super().__init__(address)
        self.data = data
        self.exists = exists

    @staticmethod
    def gql_query(address: str) -> ParamsOfQueryCollection:
        return ParamsOfQueryCollection(
            collection='accounts',
            result='boc',
            limit=1,
            filter={'id': {'eq': address}},
        )

    @classmethod
    async def from_query_result(cls, client: Client, idx: str, result: ResultOfQueryCollection) -> 'Account':
        try:
            boc = result.result[0]['boc']
        except IndexError:
            boc = ...  # None is returned for Deleted accounts (acc_type=3, acc_type_name=NonExist)
        return await cls.from_boc(client, boc, idx=idx)

    @classmethod
    async def from_boc(cls, client: Client, boc: str, **kwargs) -> 'Account':
        kw_address = kwargs.pop('idx', None)
        if boc is ...:  # account never existed
            assert kw_address is not None, 'Account must have boc or address'
            return cls(kw_address, data=None, exists=False)
        if boc is None:  # account is deleted
            assert kw_address is not None, 'Account must have boc or address'
            return cls(kw_address, data=None, exists=True)
        parse_params = ParamsOfParse(boc=boc)
        parsed = await client.boc.parse_account(params=parse_params)
        address = parsed.parsed['id']
        data = AccountData(**parsed.parsed, **kwargs)
        return cls(address, data)

    @classmethod
    async def from_address(cls, client: Client, address: str) -> 'Account':
        return await cls.from_idx(client, address)

    @property
    def address(self) -> str:
        return self.idx

    @property
    def balance(self) -> int:
        return self.data.balance



from abc import ABC, abstractmethod
from typing import Any

import attrs
from tonclient.types import ParamsOfQueryCollection, ResultOfQueryCollection

from tvmbase.client import Client
from tvmbase.utils.attrs_helper import convert_nested


class BaseTvm(ABC):
    DATA_TYPE: type = None

    def __init__(self, idx: str, data: Any = None):
        self.idx = idx
        self.data = data

    @classmethod
    async def from_idx(cls, client: Client, idx: str) -> 'BaseTvm':
        query = cls.gql_query(idx)
        result = await client.net.query_collection(params=query)
        return await cls.from_query_result(client, idx, result)

    @staticmethod
    @abstractmethod
    def gql_query(idx: str) -> ParamsOfQueryCollection:
        pass

    @classmethod
    async def from_query_result(cls, client: Client, idx: str, result: ResultOfQueryCollection) -> 'BaseTvm':
        boc = result.result[0]['boc']
        return await cls.from_boc(client, boc)

    @classmethod
    @abstractmethod
    async def from_boc(cls, client: Client, boc: str, **kwargs) -> 'BaseTvm':
        pass

    @classmethod
    def from_dict(cls, dump: dict) -> 'BaseTvm':
        idx = dump['idx']
        data = dump['data']
        if data is not None:
            convert_nested(data, cls.DATA_TYPE)
            data = cls.DATA_TYPE(**data)
        return cls(idx, data)

    def to_dict(self) -> dict:
        data = attrs.asdict(self.data) if self.data is not None else None
        return {
            'idx': self.idx,
            'data': data,
        }

    def __eq__(self, other: 'BaseTvm') -> bool:
        return self.idx == other.idx

    def __lt__(self, other: 'BaseTvm') -> bool:
        return self.idx < other.idx

    def __str__(self) -> str:
        return f'{self.__class__.__name__}<{self.idx}>'

    __repr__ = __str__



from typing import Any, TYPE_CHECKING

from tonclient.client import TonClient
from tonclient.types import (
    ClientConfig,
    NetworkConfig,
    AbiContract,
    Abi,
    CallSet,
    ParamsOfEncodeMessage,
    Signer,
    ExecutionOptions,
    ParamsOfRunTvm,
)

from tvmbase.exceptions import ClientRunLocalException
from tvmbase.models.network import Network
from tvmbase.utils.singleton import SingletonMeta

if TYPE_CHECKING:
    from tvmbase.models.tvm.account import Account


class Client(TonClient, metaclass=SingletonMeta):

    def __init__(self, network: Network):
        self.network = network
        config = self.create_config(network)
        super().__init__(config, is_async=True)

    @staticmethod
    def create_config(network: Network) -> ClientConfig:
        network_config = NetworkConfig(endpoints=network.endpoints)
        return ClientConfig(network=network_config)

    async def run_local(
            self,
            abi: AbiContract,
            method: str,
            account: 'Account' = None,
            account_params: tuple[str, str] = None,
            params: dict = None,
            auto_answer_id: bool = True,
            execution_options: dict = None,
            parse: bool = False,
    ) -> dict | Any:
        """
        Use account or address_params
        :return: dict in case of `parse` is False, otherwise parse response and return Any type
        """
        assert account is not None or account_params is not None, 'Arg `account` or `account_params` must be provided'
        if account is not None:
            address = account.address
            account_boc = account.data.boc
        else:
            address, account_boc = account_params

        if auto_answer_id:
            params = params or dict()
            self._set_answer_id(abi, method, params)

        abi = Abi.Contract(abi)
        call_set = CallSet(function_name=method, input=params)
        message_params = ParamsOfEncodeMessage(
            abi=abi,
            address=address,
            signer=Signer.NoSigner(),
            call_set=call_set,
        )
        message = await self.abi.encode_message(params=message_params)

        if execution_options is not None:
            execution_options = ExecutionOptions(**execution_options)

        run_params = ParamsOfRunTvm(
            message=message.message,
            account=account_boc,
            abi=abi,
            execution_options=execution_options,
        )
        result = await self.tvm.run_tvm(params=run_params)
        decoded = result.decoded.output

        if parse:
            if len(decoded) != 1:
                raise ClientRunLocalException('Must be exact one output param for automatic parse')
            return decoded.popitem()[1]
        else:
            return decoded

    @staticmethod
    def _set_answer_id(abi_contract: AbiContract, method: str, params: dict):
        function = next(filter(
            lambda f: f.name == method,
            abi_contract.functions
        ))
        if len(function.inputs) == 0:
            return
        first_arg_name = function.inputs[0].name
        if (first_arg_name in ('answerId', '_answer_id')) and (first_arg_name not in params):
            params[first_arg_name] = 0

    # todo
    # async def emulate_local(
    #         self,
    #         account: 'Account',
    #         abi: str,
    #         method: str,
    #         params: dict,
    #         value: int,
    #         bounce: bool = None,
    #         sender: str = None,  # todo jenkins
    # ) -> ResultOfRunTvm:
    #     # print(value / 1e9)
    #     if sender is None:
    #         sender = '0:' + '12' * 32  # todo
    #     abi = Abi.Json(abi)
    #     call_set = CallSet(function_name=method, input=params)
    #     message_params = ParamsOfEncodeInternalMessage(
    #         value=str(value),
    #         abi=abi,
    #         address=account.address,
    #         call_set=call_set,
    #         bounce=bounce,
    #         src_address=sender,
    #     )
    #     message = await self.abi.encode_internal_message(params=message_params)
    #     # run_params = ParamsOfRunTvm(
    #     #     message=message.message,
    #     #     account=account.data.boc,
    #     #     abi=abi,
    #     #     return_updated_account=True
    #     # )
    #     # return await self.tvm.run_tvm(params=run_params)
    #     account = AccountForExecutor.Account(boc=account.data.boc)
    #     run_params = ParamsOfRunExecutor(
    #         message=message.message,
    #         account=account,
    #         abi=abi,
    #         skip_transaction_check=True,
    #         return_updated_account=True,
    #     )
    #     return await self.tvm.run_executor(params=run_params)



import re

EMPTY_CELL = 'te6ccgEBAQEAAgAAAA=='  # b5ee9c72 | 01 01 01 01 | 00 02 | 00 | 00 00
EVER = 1_000_000_000  # 1e9 grams = 1 ever
HALF_EVER = 500_000_000  # 0.5 ever

# todo flags

GRAPHQL_LIMIT = 50

ADDRESS_REGEX = re.compile(r'(0|-1):[0-9a-f]{64}')
IDX_REGEX = re.compile(r'[0-9a-f]{64}')



import re

from setuptools import setup, find_packages

NAME = 'tvmbase'
URL = f'https://github.com/abionics/{NAME}'


def get_version() -> str:
    code = read_file(f'{NAME}/__init__.py')
    return re.search(r'__version__ = \'(.+?)\'', code).group(1)


def load_readme() -> str:
    return read_file('README.md')


def read_file(filename: str) -> str:
    with open(filename) as file:
        return file.read()


setup(
    name=NAME,
    version=get_version(),
    description='Part of Evertrace project | Base TVM models and utils',
    long_description=load_readme(),
    long_description_content_type='text/markdown',
    author='Alex Ermolaev',
    author_email='abionics.dev@gmail.com',
    url=URL,
    keywords='evertrace tvm utils',
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(exclude=['tests', 'examples']),
    install_requires=[
        'ton-client-py>=1.37.0.0',
        'attrs>=22.1.0',
        'loguru>=0.6.0',
        'jsonizer>=1.2.0',
    ],
    zip_safe=False,
)



import attrs


def hex_converter(value: str | int | None) -> int | None:
    if isinstance(value, int):
        return value
    elif value is None:
        return None
    else:
        return int(value, 16)


@attrs.define(frozen=True)
class _GraphQLMeta:
    json_version: int


@attrs.define(frozen=True)
class OtherCurrencyData:
    currency: int
    value: int = attrs.field(converter=hex_converter)


@attrs.define(frozen=True)
class BlockData:
    pass


@attrs.define(frozen=True)
class AccountData(_GraphQLMeta):
    id: str
    acc_type: int
    acc_type_name: str  # Uninit, Active, Frozen, NonExist
    balance: int = attrs.field(converter=hex_converter)
    bits: int = attrs.field(converter=hex_converter)
    boc: str
    cells: int = attrs.field(converter=hex_converter)
    last_paid: int
    last_trans_lt: int = attrs.field(converter=hex_converter)
    public_cells: int = attrs.field(converter=hex_converter)
    workchain_id: int
    balance_other: list[OtherCurrencyData] = None
    code: str = None
    code_hash: str = None
    data: str = None
    data_hash: str = None
    due_payment: int = attrs.field(converter=hex_converter, default=None)
    init_code_hash: str = None
    library: str = None
    library_hash: str = None
    proof: str = None
    split_depth: int = None
    state_hash: str = None
    tick: bool = None
    tock: bool = None


@attrs.define(frozen=True)
class MessageData(_GraphQLMeta):
    id: str
    boc: str
    dst: str
    msg_type: int
    msg_type_name: str  # Internal, ExtIn, ExtOut
    src: str
    status: int
    status_name: str
    block: BlockData = None
    block_id: str = None
    body: str = None
    body_hash: str = None
    bounce: bool = None
    bounced: bool = None
    chain_order: str = None
    code: str = None
    code_hash: str = None
    created_at: int = None
    created_at_string: str = None
    created_lt: int = attrs.field(converter=hex_converter, default=None)
    data: str = None
    data_hash: str = None
    dst_account: AccountData = None
    dst_transaction: 'TransactionData' = None
    dst_workchain_id: int = None
    fwd_fee: int = attrs.field(converter=hex_converter, default=None)
    ihr_disabled: bool = None
    ihr_fee: int = attrs.field(converter=hex_converter, default=None)
    import_fee: int = attrs.field(converter=hex_converter, default=None)
    library: str = None
    library_hash: str = None
    proof: str = None
    split_depth: int = None
    src_account: AccountData = None
    src_transaction: 'TransactionData' = None
    src_workchain_id: int = None
    tick: bool = None
    tock: bool = None
    value: int = attrs.field(converter=hex_converter, default=None)
    value_other: list[OtherCurrencyData] = None
    dst_transaction_id: str = None  # not exists in GraphQL
    src_transaction_id: str = None  # not exists in GraphQL


@attrs.define(frozen=True)
class TransactionActionData:
    action_list_hash: str
    msgs_created: int
    no_funds: bool
    result_code: int
    skipped_actions: int
    spec_actions: int
    status_change: int
    success: bool
    tot_actions: int
    tot_msg_size_bits: int
    tot_msg_size_cells: int
    valid: bool
    result_arg: int = None
    status_change_name: str = None
    total_action_fees: int = attrs.field(converter=hex_converter, default=None)
    total_fwd_fees: int = attrs.field(converter=hex_converter, default=None)


@attrs.define(frozen=True)
class TransactionBounceData:
    bounce_type: int
    bounce_type_name: str
    msg_size_bits: int
    msg_size_cells: int
    fwd_fees: int = attrs.field(converter=hex_converter, default=None)
    msg_fees: int = attrs.field(converter=hex_converter, default=None)
    req_fwd_fees: int = attrs.field(converter=hex_converter, default=None)


@attrs.define(frozen=True)
class TransactionComputeData:
    compute_type: int
    compute_type_name: str
    account_activated: bool = None
    exit_arg: int = None
    exit_code: int = None
    gas_credit: int = None
    gas_fees: int = attrs.field(converter=hex_converter, default=None)
    gas_limit: int = None
    gas_used: int = None
    mode: int = None
    msg_state_used: bool = None
    skipped_reason: int = None
    skipped_reason_name: str = None
    success: bool = None
    vm_final_state_hash: str = None
    vm_init_state_hash: str = None
    vm_steps: int = None


@attrs.define(frozen=True)
class TransactionCreditData:
    credit: int = attrs.field(converter=hex_converter)
    due_fees_collected: int = attrs.field(converter=hex_converter, default=None)
    credit_other: list[OtherCurrencyData] = None


@attrs.define(frozen=True)
class TransactionSplitInfoData:
    acc_split_depth: int
    cur_shard_pfx_len: int
    sibling_addr: str
    this_addr: str


@attrs.define(frozen=True)
class TransactionStorageData:
    status_change: int
    status_change_name: str
    storage_fees_collected: int = attrs.field(converter=hex_converter)
    storage_fees_due: int = attrs.field(converter=hex_converter, default=None)


@attrs.define(frozen=True)
class TransactionData(_GraphQLMeta):
    id: str
    aborted: bool
    account_addr: str
    balance_delta: int = attrs.field(converter=hex_converter)
    boc: str
    compute: TransactionComputeData
    credit_first: bool
    destroyed: bool
    end_status: int
    end_status_name: str
    in_msg: str
    lt: int = attrs.field(converter=hex_converter)
    new_hash: str
    now: int
    old_hash: str
    orig_status: int
    orig_status_name: str
    out_msgs: list[str]
    outmsg_cnt: int
    prev_trans_hash: str
    prev_trans_lt: int = attrs.field(converter=hex_converter)
    status: int
    status_name: str
    total_fees: int = attrs.field(converter=hex_converter)
    tr_type: int
    tr_type_name: str
    workchain_id: int
    account: AccountData = None
    action: TransactionActionData = None
    balance_delta_other: list[OtherCurrencyData] = None
    block: BlockData = None
    block_id: str = None
    bounce: TransactionBounceData = None
    chain_order: str = None
    credit: TransactionCreditData = None
    ext_in_msg_fee: int = attrs.field(converter=hex_converter, default=None)
    in_message: MessageData = None
    installed: bool = None
    now_string: str = None
    out_messages: list[MessageData] = None
    prepare_transaction: str = None
    proof: str = None
    split_info: TransactionSplitInfoData = None
    storage: TransactionStorageData = None
    total_fees_other: list[OtherCurrencyData] = None
    tt: str = None
